from typing import Any, Optional, Dict, TypedDict, Literal
from datetime import datetime


class UsageMetadata(TypedDict, total=False):
    promptTokens: Optional[int]
    completionTokens: Optional[int]
    totalTokens: Optional[int]


class TrackedRequest(TypedDict, total=False):
    id: str
    project_name: Optional[str]
    timestamp: datetime
    response_time: Optional[float]
    model: str
    prompt: Any
    response: Any
    usage: Optional[UsageMetadata]
    error: Optional[Any]
    search_grounding: Optional[bool]
    received_search_grounded_response: Optional[bool]
    metadata: Optional[Dict[str, Any]]


class StorageOptions(TypedDict, total=False):
    file_path: Optional[str]
    auto_save: Optional[bool]


class RemoteStorageOptions(TypedDict, total=False):
    endpoint: str
    api_key: str
    project_name: Optional[str]
    timeout: Optional[int]
    retries: Optional[int]
    retry_delay: Optional[int]


StorageMode = Literal["local", "remote"]


class TrackerOptions(TypedDict, total=False):
    api_key: Optional[str]
    vertexai: Optional[bool]
    project: Optional[str]
    location: Optional[str]
    storage: Optional[StorageMode]
    local_storage: Optional[StorageOptions]
    remote_storage: Optional[RemoteStorageOptions]



from .tracker import Client
from .types import TrackedRequest, TrackerOptions, StorageOptions, RemoteStorageOptions, StorageMode
from . import genai

__version__ = "0.1.0"
__all__ = ["Client", "genai", "TrackedRequest", "TrackerOptions", "StorageOptions", "RemoteStorageOptions", "StorageMode"]

GeminiTracker = Client



import json
import os
from pathlib import Path
from typing import List, Optional, Protocol
from .types import TrackedRequest, StorageOptions


class IStorage(Protocol):
    def add(self, request: TrackedRequest) -> None:
        ...

    def get_all(self) -> List[TrackedRequest]:
        ...

    def clear(self) -> None:
        ...


class LocalStorage:
    def __init__(self, options: Optional[StorageOptions] = None):
        options = options or {}
        self.file_path = options.get("file_path", "./gemini-tracker-logs.json")
        self.auto_save = options.get("auto_save", True)
        self.requests: List[TrackedRequest] = []
        self.load()

    def add(self, request: TrackedRequest) -> None:
        self.requests.append(request)
        if self.auto_save:
            self.save()

    def get_all(self) -> List[TrackedRequest]:
        return self.requests.copy()

    def save(self) -> None:
        path = Path(self.file_path)
        path.parent.mkdir(parents=True, exist_ok=True)

        serializable_requests = []
        for req in self.requests:
            req_copy = req.copy()
            if "timestamp" in req_copy and hasattr(req_copy["timestamp"], "isoformat"):
                req_copy["timestamp"] = req_copy["timestamp"].isoformat()
            serializable_requests.append(req_copy)

        with open(self.file_path, "w") as f:
            json.dump(serializable_requests, f, indent=2, default=str)

    def load(self) -> None:
        if os.path.exists(self.file_path):
            with open(self.file_path, "r") as f:
                self.requests = json.load(f)

    def clear(self) -> None:
        self.requests = []



import time
import traceback
from typing import List, Optional
import httpx
from .types import TrackedRequest, RemoteStorageOptions


class RemoteStorage:
    def __init__(self, options: RemoteStorageOptions):
        if not options.get("api_key"):
            raise ValueError("remoteStorage.api_key is required for remote storage")

        self.endpoint = options.get("endpoint", "https://api.tokentracker.dev/api/track")
        self.api_key = options["api_key"]
        self.project_name = options.get("project_name")
        self.timeout = options.get("timeout", 10000) / 1000
        self.retries = options.get("retries", 3)
        self.retry_delay = options.get("retry_delay", 1000) / 1000

    def _serialize_error(self, error: Exception) -> dict:
        serialized = {
            "name": type(error).__name__,
            "message": str(error),
            "stack": "".join(traceback.format_exception(type(error), error, error.__traceback__)),
        }

        for key in dir(error):
            if not key.startswith("_"):
                try:
                    value = getattr(error, key)
                    if not callable(value):
                        serialized[key] = value
                except:
                    pass

        error_msg = str(error)
        if error_msg.strip().startswith(("{", "[")):
            try:
                import json
                serialized["parsedMessage"] = json.loads(error_msg)
            except:
                pass

        return serialized

    def _fetch_with_retry(
        self, url: str, method: str, headers: dict, json_data: Optional[dict] = None, attempt: int = 0
    ) -> httpx.Response:
        try:
            with httpx.Client(timeout=self.timeout, follow_redirects=True) as client:
                if method == "POST":
                    response = client.post(url, headers=headers, json=json_data)
                else:
                    response = client.get(url, headers=headers)

                if not response.is_success and response.status_code >= 500 and attempt < self.retries:
                    time.sleep(self.retry_delay * (2 ** attempt))
                    return self._fetch_with_retry(url, method, headers, json_data, attempt + 1)

                return response
        except (httpx.TimeoutException, httpx.ConnectError) as error:
            if attempt < self.retries:
                time.sleep(self.retry_delay * (2 ** attempt))
                return self._fetch_with_retry(url, method, headers, json_data, attempt + 1)
            raise error

    def add(self, request: TrackedRequest) -> None:
        request_with_project = {
            **request,
            "projectName": self.project_name,
            "responseTime": request.get("response_time"),
            "searchGrounding": request.get("search_grounding", False),
            "receivedSearchGroundedResponse": request.get("received_search_grounded_response", False),
            "error": self._serialize_error(request["error"]) if request.get("error") else None,
            "metadata": request.get("metadata"),
        }

        if "timestamp" in request_with_project and hasattr(request_with_project["timestamp"], "isoformat"):
            request_with_project["timestamp"] = request_with_project["timestamp"].isoformat()

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }

        response = self._fetch_with_retry(
            self.endpoint,
            "POST",
            headers,
            request_with_project
        )

        if not response.is_success:
            raise Exception(f"Remote storage failed: {response.status_code} {response.reason_phrase}")

    def get_all(self) -> List[TrackedRequest]:
        headers = {
            "Authorization": f"Bearer {self.api_key}",
        }

        response = self._fetch_with_retry(self.endpoint, "GET", headers)

        if not response.is_success:
            raise Exception(f"Failed to fetch tracked requests: {response.status_code}")

        return response.json()

    def clear(self) -> None:
        raise Exception("Clear not supported for remote storage")



import uuid
import time
from datetime import datetime
from typing import Any, Optional, Iterator
from google import genai
from .storage import LocalStorage, IStorage
from .remote_storage import RemoteStorage
from .types import TrackedRequest, TrackerOptions


class Client:
    def __init__(
        self,
        api_key: Optional[str] = None,
        vertexai: bool = False,
        project: Optional[str] = None,
        location: Optional[str] = None,
        storage: str = "local",
        local_storage: Optional[dict] = None,
        remote_storage: Optional[dict] = None,
        **kwargs
    ):
        storage_mode = storage
        local_storage_options = local_storage
        remote_storage_options = remote_storage

        client_kwargs = {}
        if api_key:
            client_kwargs["api_key"] = api_key
        if vertexai:
            client_kwargs["vertexai"] = True
            if project:
                client_kwargs["project"] = project
            if location:
                client_kwargs["location"] = location

        self._client = genai.Client(**client_kwargs)

        if storage_mode == "remote":
            if not remote_storage_options:
                raise ValueError("remote_storage options required when storage mode is 'remote'")
            self._storage: IStorage = RemoteStorage(remote_storage_options)
        else:
            self._storage: IStorage = LocalStorage(local_storage_options)
        self._original_generate_content = self._client.models.generate_content
        self._original_generate_content_stream = self._client.models.generate_content_stream

        self._client.models.generate_content = self._wrapped_generate_content
        self._client.models.generate_content_stream = self._wrapped_generate_content_stream

    def _wrapped_generate_content(self, **params: Any) -> Any:
        request_id = str(uuid.uuid4())
        timestamp = datetime.now()
        start_time = time.perf_counter()

        metadata = params.pop("metadata", None)
        model = params.get("model", "unknown")
        contents = params.get("contents")

        config = params.get("config")
        tools = config.tools if config and hasattr(config, "tools") else params.get("tools")
        search_grounding = bool(
            tools and hasattr(tools, "__iter__") and
            any(hasattr(tool, "google_search") for tool in tools)
        )

        try:
            response = self._original_generate_content(**params)
            response_time = (time.perf_counter() - start_time) * 1000

            received_search_grounded_response = bool(
                hasattr(response, "candidates") and
                response.candidates and
                hasattr(response.candidates[0], "grounding_metadata") and
                response.candidates[0].grounding_metadata and
                (hasattr(response.candidates[0].grounding_metadata, "search_entry_point") or
                 hasattr(response.candidates[0].grounding_metadata, "grounding_chunks"))
            )

            usage_metadata = response.usage_metadata if hasattr(response, "usage_metadata") else None

            response_dict = self._serialize_response(response)

            tracked_request: TrackedRequest = {
                "id": request_id,
                "timestamp": timestamp,
                "response_time": response_time,
                "model": model,
                "prompt": contents,
                "response": response_dict,
                "usage": {
                    "promptTokens": usage_metadata.prompt_token_count if usage_metadata else None,
                    "completionTokens": usage_metadata.candidates_token_count if usage_metadata else None,
                    "totalTokens": usage_metadata.total_token_count if usage_metadata else None,
                },
                "search_grounding": search_grounding,
                "received_search_grounded_response": received_search_grounded_response,
                "metadata": metadata,
            }

            try:
                self._storage.add(tracked_request)
            except Exception as storage_error:
                print(f"Failed to save tracked request: {storage_error}")

            return response

        except Exception as error:
            tracked_request: TrackedRequest = {
                "id": request_id,
                "timestamp": timestamp,
                "model": model,
                "prompt": contents,
                "response": None,
                "error": str(error),
                "search_grounding": search_grounding,
                "received_search_grounded_response": False,
                "metadata": metadata,
            }

            try:
                self._storage.add(tracked_request)
            except Exception as storage_error:
                print(f"Failed to save tracked request with error: {storage_error}")

            raise

    def _serialize_response(self, response: Any) -> dict:
        result = {}

        if hasattr(response, "sdk_http_response"):
            sdk_resp = response.sdk_http_response
            result["sdkHttpResponse"] = {
                "headers": dict(sdk_resp.headers) if hasattr(sdk_resp, "headers") else {}
            }

        if hasattr(response, "candidates"):
            result["candidates"] = []
            for candidate in response.candidates:
                cand_dict = {}
                if hasattr(candidate, "content"):
                    content = candidate.content
                    cand_dict["content"] = {
                        "parts": [{"text": part.text} for part in content.parts if hasattr(part, "text")],
                        "role": content.role if hasattr(content, "role") else None
                    }
                if hasattr(candidate, "finish_reason"):
                    cand_dict["finishReason"] = str(candidate.finish_reason).replace("FinishReason.", "")
                if hasattr(candidate, "grounding_metadata"):
                    cand_dict["groundingMetadata"] = {}
                if hasattr(candidate, "index"):
                    cand_dict["index"] = candidate.index
                result["candidates"].append(cand_dict)

        if hasattr(response, "model_version"):
            result["modelVersion"] = response.model_version
        if hasattr(response, "response_id"):
            result["responseId"] = response.response_id

        if hasattr(response, "usage_metadata"):
            um = response.usage_metadata
            result["usageMetadata"] = {
                "promptTokenCount": getattr(um, "prompt_token_count", None),
                "candidatesTokenCount": getattr(um, "candidates_token_count", None),
                "totalTokenCount": getattr(um, "total_token_count", None),
            }
            if hasattr(um, "prompt_tokens_details"):
                result["usageMetadata"]["promptTokensDetails"] = [
                    {"modality": str(d.modality).replace("MediaModality.", ""), "tokenCount": d.token_count}
                    for d in um.prompt_tokens_details
                ]
            if hasattr(um, "thoughts_token_count"):
                result["usageMetadata"]["thoughtsTokenCount"] = um.thoughts_token_count

        return result

    def _wrapped_generate_content_stream(self, **params: Any) -> Iterator[Any]:
        request_id = str(uuid.uuid4())
        timestamp = datetime.now()
        start_time = time.perf_counter()

        metadata = params.pop("metadata", None)
        model = params.get("model", "unknown")
        contents = params.get("contents")

        config = params.get("config")
        tools = config.tools if config and hasattr(config, "tools") else params.get("tools")
        search_grounding = bool(
            tools and hasattr(tools, "__iter__") and
            any(hasattr(tool, "google_search") for tool in tools)
        )

        try:
            stream = self._original_generate_content_stream(**params)
            chunks = []
            full_response = None

            for chunk in stream:
                chunks.append(chunk)
                yield chunk

            if chunks:
                full_response = chunks[-1]

            response_time = (time.perf_counter() - start_time) * 1000

            received_search_grounded_response = bool(
                full_response and
                hasattr(full_response, "candidates") and
                full_response.candidates and
                hasattr(full_response.candidates[0], "grounding_metadata") and
                full_response.candidates[0].grounding_metadata and
                (hasattr(full_response.candidates[0].grounding_metadata, "search_entry_point") or
                 hasattr(full_response.candidates[0].grounding_metadata, "grounding_chunks"))
            )

            usage_metadata = full_response.usage_metadata if full_response and hasattr(full_response, "usage_metadata") else None

            response_dict = self._serialize_response(full_response) if full_response else None

            tracked_request: TrackedRequest = {
                "id": request_id,
                "timestamp": timestamp,
                "response_time": response_time,
                "model": model,
                "prompt": contents,
                "response": response_dict,
                "usage": {
                    "promptTokens": usage_metadata.prompt_token_count if usage_metadata else None,
                    "completionTokens": usage_metadata.candidates_token_count if usage_metadata else None,
                    "totalTokens": usage_metadata.total_token_count if usage_metadata else None,
                },
                "search_grounding": search_grounding,
                "received_search_grounded_response": received_search_grounded_response,
                "metadata": metadata,
            }

            try:
                self._storage.add(tracked_request)
            except Exception as storage_error:
                print(f"Failed to save streamed tracked request: {storage_error}")

        except Exception as error:
            tracked_request: TrackedRequest = {
                "id": request_id,
                "timestamp": timestamp,
                "model": model,
                "prompt": contents,
                "response": None,
                "error": str(error),
                "search_grounding": search_grounding,
                "received_search_grounded_response": False,
                "metadata": metadata,
            }

            try:
                self._storage.add(tracked_request)
            except Exception as storage_error:
                print(f"Failed to save streamed tracked request with error: {storage_error}")

            raise

    @property
    def models(self):
        return self._client.models

    @property
    def chats(self):
        return self._client.chats

    @property
    def files(self):
        return self._client.files

    @property
    def caches(self):
        return self._client.caches

    def get_tracked_requests(self) -> list[TrackedRequest]:
        return self._storage.get_all()

    def load_history(self) -> None:
        if isinstance(self._storage, LocalStorage):
            self._storage.load()

    def save_history(self) -> None:
        if isinstance(self._storage, LocalStorage):
            self._storage.save()

    def clear_history(self) -> None:
        self._storage.clear()
